Updated asyncns.[ch]
authorMikael Hallendal <micke@imendio.com>
Sun, 26 Oct 2008 01:52:23 +0200
changeset 549 99a387fdb1bb
parent 548 60c1d6cbde10
child 550 104d37c1b24a
Updated asyncns.[ch]
loudmouth/asyncns.c
loudmouth/asyncns.h
loudmouth/lm-old-socket.c
loudmouth/lm-resolver.c
--- a/loudmouth/asyncns.c	Sat Oct 25 22:33:23 2008 +0200
+++ b/loudmouth/asyncns.c	Sun Oct 26 01:52:23 2008 +0200
@@ -1,28 +1,29 @@
-/* -*- mode: c; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
-/* $Id: asyncns.c 27 2007-02-16 13:51:03Z lennart $ */
+/***
+  This file is part of libasyncns.
+
+  Copyright 2005-2008 Lennart Poettering
 
-/***
-    This file is part of libasyncns.
- 
-    libasyncns is free software; you can redistribute it and/or modify
-    it under the terms of the GNU Lesser General Public License as
-    published by the Free Software Foundation; either version 2 of the
-    License, or (at your option) any later version.
- 
-    libasyncns is distributed in the hope that it will be useful, but
-    WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
-    General Public License for more details.
- 
-    You should have received a copy of the GNU Lesser General Public
-    License along with libasyncns; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
-    USA.
+  libasyncns is free software; you can redistribute it and/or modify
+  it under the terms of the GNU Lesser General Public License as
+  published by the Free Software Foundation, either version 2.1 of the
+  License, or (at your option) any later version.
+
+  libasyncns is distributed in the hope that it will be useful, but
+  WITHOUT ANY WARRANTY; without even the implied warranty of
+  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
+  Lesser General Public License for more details.
+
+  You should have received a copy of the GNU Lesser General Public
+  License along with libasyncns. If not, see
+  <http://www.gnu.org/licenses/>.
 ***/
 
-/*#undef HAVE_PTHREAD */
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
 
-#include <glib.h>
+/* #undef HAVE_PTHREAD */
+
 #include <assert.h>
 #include <fcntl.h>
 #include <signal.h>
@@ -38,6 +39,7 @@
 #include <netinet/in.h>
 #include <arpa/nameser.h>
 #include <resolv.h>
+#include <dirent.h>
 
 #ifdef HAVE_SYS_PRCTL_H
 #include <sys/prctl.h>
@@ -61,20 +63,22 @@
     REQUEST_RES_QUERY,
     REQUEST_RES_SEARCH,
     RESPONSE_RES,
-    REQUEST_TERMINATE
+    REQUEST_TERMINATE,
+    RESPONSE_DIED
 } query_type_t;
 
 enum {
     REQUEST_RECV_FD = 0,
     REQUEST_SEND_FD = 1,
     RESPONSE_RECV_FD = 2,
-    RESPONSE_SEND_FD = 3
+    RESPONSE_SEND_FD = 3,
+    MESSAGE_FD_MAX = 4
 };
 
 struct asyncns {
     int fds[4];
 
-#ifndef HAVE_PTHREAD    
+#ifndef HAVE_PTHREAD
     pid_t workers[MAX_WORKERS];
 #else
     pthread_t workers[MAX_WORKERS];
@@ -87,6 +91,7 @@
     asyncns_query_t *done_head, *done_tail;
 
     int n_queries;
+    int dead;
 };
 
 struct asyncns_query {
@@ -96,6 +101,8 @@
     query_type_t type;
     asyncns_query_t *done_next, *done_prev;
     int ret;
+    int _errno;
+    int _h_errno;
     struct addrinfo *addrinfo;
     char *serv, *host;
     void *userdata;
@@ -120,6 +127,9 @@
 typedef struct addrinfo_response {
     struct rheader header;
     int ret;
+    int _errno;
+    int _h_errno;
+    /* followed by addrinfo_serialization[] */
 } addrinfo_response_t;
 
 typedef struct addrinfo_serialization {
@@ -129,6 +139,7 @@
     int ai_protocol;
     size_t ai_addrlen;
     size_t canonname_len;
+    /* Followed by ai_addr amd ai_canonname with variable lengths */
 } addrinfo_serialization_t;
 
 typedef struct nameinfo_request {
@@ -142,20 +153,199 @@
     struct rheader header;
     size_t hostlen, servlen;
     int ret;
+    int _errno;
+    int _h_errno;
 } nameinfo_response_t;
 
 typedef struct res_query_request {
     struct rheader header;
     int class;
     int type;
-    size_t dlen;
+    size_t dname_len;
 } res_request_t;
 
 typedef struct res_query_response {
     struct rheader header;
     int ret;
+    int _errno;
+    int _h_errno;
 } res_response_t;
 
+#ifndef HAVE_STRNDUP
+
+static char *strndup(const char *s, size_t l) {
+    size_t a;
+    char *n;
+
+    a = strlen(s);
+    if (a > l)
+        a = l;
+
+    if (!(n = malloc(a+1)))
+        return NULL;
+
+    memcpy(n, s, a);
+    n[a] = 0;
+
+    return n;
+}
+
+#endif
+
+#ifndef HAVE_PTHREAD
+
+static int close_allv(const int except_fds[]) {
+    struct rlimit rl;
+    int fd;
+    int saved_errno;
+
+#ifdef __linux__
+
+    DIR *d;
+
+    assert(except_fds);
+
+    if ((d = opendir("/proc/self/fd"))) {
+
+        struct dirent *de;
+
+        while ((de = readdir(d))) {
+            int found;
+            long l;
+            char *e = NULL;
+            int i;
+
+            if (de->d_name[0] == '.')
+                continue;
+
+            errno = 0;
+            l = strtol(de->d_name, &e, 10);
+            if (errno != 0 || !e || *e) {
+                closedir(d);
+                errno = EINVAL;
+                return -1;
+            }
+
+            fd = (int) l;
+
+            if ((long) fd != l) {
+                closedir(d);
+                errno = EINVAL;
+                return -1;
+            }
+
+            if (fd < 3)
+                continue;
+
+            if (fd == dirfd(d))
+                continue;
+
+            found = 0;
+            for (i = 0; except_fds[i] >= 0; i++)
+                if (except_fds[i] == fd) {
+                    found = 1;
+                    break;
+                }
+
+            if (found)
+                continue;
+
+            if (close(fd) < 0) {
+                saved_errno = errno;
+                closedir(d);
+                errno = saved_errno;
+
+                return -1;
+            }
+        }
+
+        closedir(d);
+        return 0;
+    }
+
+#endif
+
+    if (getrlimit(RLIMIT_NOFILE, &rl) < 0)
+        return -1;
+
+    for (fd = 0; fd < (int) rl.rlim_max; fd++) {
+        int i;
+
+        if (fd <= 3)
+            continue;
+
+        for (i = 0; except_fds[i] >= 0; i++)
+            if (except_fds[i] == fd)
+                continue;
+
+        if (close(fd) < 0 && errno != EBADF)
+            return -1;
+    }
+
+    return 0;
+}
+
+static int reset_sigsv(const int except[]) {
+    int sig;
+    assert(except);
+
+    for (sig = 1; sig < _NSIG; sig++) {
+        int reset = 1;
+
+        switch (sig) {
+            case SIGKILL:
+            case SIGSTOP:
+                reset = 0;
+                break;
+
+            default: {
+                int i;
+
+                for (i = 0; except[i] > 0; i++) {
+                    if (sig == except[i]) {
+                        reset = 0;
+                        break;
+                    }
+                }
+            }
+        }
+
+        if (reset) {
+            struct sigaction sa;
+
+            memset(&sa, 0, sizeof(sa));
+            sa.sa_handler = SIG_DFL;
+
+            /* On Linux the first two RT signals are reserved by
+             * glibc, and sigaction() will return EINVAL for them. */
+            if ((sigaction(sig, &sa, NULL) < 0))
+                if (errno != EINVAL)
+                    return -1;
+        }
+    }
+
+    return 0;
+}
+
+static int ignore_sigsv(const int ignore[]) {
+    int i;
+    assert(ignore);
+
+    for (i = 0; ignore[i] > 0; i++) {
+        struct sigaction sa;
+
+        memset(&sa, 0, sizeof(sa));
+        sa.sa_handler = SIG_IGN;
+
+        if ((sigaction(ignore[i], &sa, NULL) < 0))
+            return -1;
+    }
+
+    return 0;
+}
+
+#endif
+
 static int fd_nonblock(int fd) {
     int i;
     assert(fd >= 0);
@@ -178,13 +368,24 @@
 
     if (v & FD_CLOEXEC)
         return 0;
-    
+
     return fcntl(fd, F_SETFD, v | FD_CLOEXEC);
 }
 
+static int send_died(int out_fd) {
+    rheader_t rh;
+    assert(out_fd > 0);
+
+    memset(&rh, 0, sizeof(rh));
+    rh.type = RESPONSE_DIED;
+    rh.id = 0;
+    rh.length = sizeof(rh);
+
+    return send(out_fd, &rh, rh.length, 0);
+}
 
 static void *serialize_addrinfo(void *p, const struct addrinfo *ai, size_t *length, size_t maxlength) {
-    addrinfo_serialization_t *s = p;
+    addrinfo_serialization_t s;
     size_t cnl, l;
     assert(p);
     assert(ai);
@@ -197,13 +398,14 @@
     if (*length + l > maxlength)
         return NULL;
 
-    s->ai_flags = ai->ai_flags;
-    s->ai_family = ai->ai_family;
-    s->ai_socktype = ai->ai_socktype;
-    s->ai_protocol = ai->ai_protocol;
-    s->ai_addrlen = ai->ai_addrlen;
-    s->canonname_len = cnl;
+    s.ai_flags = ai->ai_flags;
+    s.ai_family = ai->ai_family;
+    s.ai_socktype = ai->ai_socktype;
+    s.ai_protocol = ai->ai_protocol;
+    s.ai_addrlen = ai->ai_addrlen;
+    s.canonname_len = cnl;
 
+    memcpy((uint8_t*) p, &s, sizeof(addrinfo_serialization_t));
     memcpy((uint8_t*) p + sizeof(addrinfo_serialization_t), ai->ai_addr, ai->ai_addrlen);
 
     if (ai->ai_canonname)
@@ -213,22 +415,29 @@
     return (uint8_t*) p + l;
 }
 
-static int send_addrinfo_reply(int out_fd, unsigned id, int ret, struct addrinfo *ai) {
-    uint8_t data[BUFSIZE];
-    addrinfo_response_t *resp = (addrinfo_response_t*) data;
+static int send_addrinfo_reply(int out_fd, unsigned id, int ret, struct addrinfo *ai, int _errno, int _h_errno) {
+    addrinfo_response_t data[BUFSIZE/sizeof(addrinfo_response_t) + 1];
+    addrinfo_response_t *resp = data;
     assert(out_fd >= 0);
 
+    memset(data, 0, sizeof(data));
     resp->header.type = RESPONSE_ADDRINFO;
     resp->header.id = id;
     resp->header.length = sizeof(addrinfo_response_t);
     resp->ret = ret;
+    resp->_errno = _errno;
+    resp->_h_errno = _h_errno;
 
     if (ret == 0 && ai) {
-        void *p = data + sizeof(addrinfo_response_t);
+        void *p = data + 1;
+        struct addrinfo *k;
+
+        for (k = ai; k; k = k->ai_next) {
 
-        while (ai && p) {
-            p = serialize_addrinfo(p, ai, &resp->header.length, BUFSIZE);
-            ai = ai->ai_next;
+            if (!(p = serialize_addrinfo(p, k, &resp->header.length, (char*) data + BUFSIZE - (char*) p))) {
+                resp->ret = EAI_MEMORY;
+                break;
+            }
         }
     }
 
@@ -238,49 +447,55 @@
     return send(out_fd, resp, resp->header.length, 0);
 }
 
-static int send_nameinfo_reply(int out_fd, unsigned id, int ret, const char *host, const char *serv) {
-    uint8_t data[BUFSIZE];
+static int send_nameinfo_reply(int out_fd, unsigned id, int ret, const char *host, const char *serv, int _errno, int _h_errno) {
+    nameinfo_response_t data[BUFSIZE/sizeof(nameinfo_response_t) + 1];
     size_t hl, sl;
-    nameinfo_response_t *resp = (nameinfo_response_t*) data;
+    nameinfo_response_t *resp = data;
 
     assert(out_fd >= 0);
-    
+
     sl = serv ? strlen(serv)+1 : 0;
     hl = host ? strlen(host)+1 : 0;
 
+    memset(data, 0, sizeof(data));
     resp->header.type = RESPONSE_NAMEINFO;
     resp->header.id = id;
     resp->header.length = sizeof(nameinfo_response_t) + hl + sl;
     resp->ret = ret;
+    resp->_errno = _errno;
+    resp->_h_errno = _h_errno;
     resp->hostlen = hl;
     resp->servlen = sl;
 
     assert(sizeof(data) >= resp->header.length);
-    
+
     if (host)
-        memcpy(data + sizeof(nameinfo_response_t), host, hl);
+        memcpy((uint8_t *)data + sizeof(nameinfo_response_t), host, hl);
 
     if (serv)
-        memcpy(data + sizeof(nameinfo_response_t) + hl, serv, sl);
-    
+        memcpy((uint8_t *)data + sizeof(nameinfo_response_t) + hl, serv, sl);
+
     return send(out_fd, resp, resp->header.length, 0);
 }
 
-static int send_res_reply(int out_fd, unsigned id, const unsigned char *answer, int ret) {
-    uint8_t data[BUFSIZE];
-    res_response_t *resp = (res_response_t *) data;
+static int send_res_reply(int out_fd, unsigned id, const unsigned char *answer, int ret, int _errno, int _h_errno) {
+    res_response_t data[BUFSIZE/sizeof(res_response_t) + 1];
+    res_response_t *resp = data;
 
     assert(out_fd >= 0);
-    
+
+    memset(data, 0, sizeof(data));
     resp->header.type = RESPONSE_RES;
     resp->header.id = id;
     resp->header.length = sizeof(res_response_t) + (ret < 0 ? 0 : ret);
-    resp->ret = (ret < 0) ? -errno : ret;
+    resp->ret = ret;
+    resp->_errno = _errno;
+    resp->_h_errno = _h_errno;
 
     assert(sizeof(data) >= resp->header.length);
-    
+
     if (ret > 0)
-        memcpy(data + sizeof(res_response_t), answer, ret);
+        memcpy((uint8_t *)data + sizeof(res_response_t), answer, ret);
 
     return send(out_fd, resp, resp->header.length, 0);
 }
@@ -292,86 +507,80 @@
     assert(length == req->length);
 
     switch (req->type) {
-    case REQUEST_ADDRINFO: {
-        struct addrinfo ai, *result = NULL;
-        const addrinfo_request_t *ai_req = (const addrinfo_request_t*) req;
-        const char *node, *service;
-        int ret;
+        case REQUEST_ADDRINFO: {
+            struct addrinfo ai, *result = NULL;
+            const addrinfo_request_t *ai_req = (const addrinfo_request_t*) req;
+            const char *node, *service;
+            int ret;
 
-        assert(length >= sizeof(addrinfo_request_t));
-        assert(length == sizeof(addrinfo_request_t) + ai_req->node_len + ai_req->service_len);
+            assert(length >= sizeof(addrinfo_request_t));
+            assert(length == sizeof(addrinfo_request_t) + ai_req->node_len + ai_req->service_len);
 
-        memset(&ai, 0, sizeof(ai));
-        ai.ai_flags = ai_req->ai_flags;
-        ai.ai_family = ai_req->ai_family;
-        ai.ai_socktype = ai_req->ai_socktype;
-        ai.ai_protocol = ai_req->ai_protocol;
+            memset(&ai, 0, sizeof(ai));
+            ai.ai_flags = ai_req->ai_flags;
+            ai.ai_family = ai_req->ai_family;
+            ai.ai_socktype = ai_req->ai_socktype;
+            ai.ai_protocol = ai_req->ai_protocol;
 
-        node = ai_req->node_len ? (const char*) req + sizeof(addrinfo_request_t) : NULL;
-        service = ai_req->service_len ? (const char*) req + sizeof(addrinfo_request_t) + ai_req->node_len : NULL;
+            node = ai_req->node_len ? (const char*) req + sizeof(addrinfo_request_t) : NULL;
+            service = ai_req->service_len ? (const char*) req + sizeof(addrinfo_request_t) + ai_req->node_len : NULL;
 
-        /*printf("[getaddrinfo for '%s']\n", node);*/
-        ret = getaddrinfo(node, service,
-                          ai_req->hints_is_null ? NULL : &ai,
-                          &result);
-        /*printf("[return value: %d: '%s']\n", ret, gai_strerror(ret));*/
+            ret = getaddrinfo(node, service,
+                              ai_req->hints_is_null ? NULL : &ai,
+                              &result);
+
+            /* send_addrinfo_reply() frees result */
+            return send_addrinfo_reply(out_fd, req->id, ret, result, errno, h_errno);
+        }
 
-        /* send_addrinfo_reply() frees result */
-        return send_addrinfo_reply(out_fd, req->id, ret, result);
-    }
+        case REQUEST_NAMEINFO: {
+            int ret;
+            const nameinfo_request_t *ni_req = (const nameinfo_request_t*) req;
+            char hostbuf[NI_MAXHOST], servbuf[NI_MAXSERV];
+            struct sockaddr_storage sa;
 
-    case REQUEST_NAMEINFO: {
-        int ret;
-        const nameinfo_request_t *ni_req = (const nameinfo_request_t*) req;
-        char hostbuf[NI_MAXHOST], servbuf[NI_MAXSERV];
-        const struct sockaddr *sa;
-            
-        assert(length >= sizeof(nameinfo_request_t));
-        assert(length == sizeof(nameinfo_request_t) + ni_req->sockaddr_len);
+            assert(length >= sizeof(nameinfo_request_t));
+            assert(length == sizeof(nameinfo_request_t) + ni_req->sockaddr_len);
+
+            memcpy(&sa, (const uint8_t *)req + sizeof(nameinfo_request_t), ni_req->sockaddr_len);
 
-        sa = (const struct sockaddr*) ((const char*) req + sizeof(nameinfo_request_t));
-            
-        ret = getnameinfo(sa, ni_req->sockaddr_len,
-                          ni_req->gethost ? hostbuf : NULL, ni_req->gethost ? sizeof(hostbuf) : 0,
-                          ni_req->getserv ? servbuf : NULL, ni_req->getserv ? sizeof(servbuf) : 0,
-                          ni_req->flags);
+            ret = getnameinfo((struct sockaddr *)&sa, ni_req->sockaddr_len,
+                              ni_req->gethost ? hostbuf : NULL, ni_req->gethost ? sizeof(hostbuf) : 0,
+                              ni_req->getserv ? servbuf : NULL, ni_req->getserv ? sizeof(servbuf) : 0,
+                              ni_req->flags);
 
-        return send_nameinfo_reply(out_fd, req->id, ret,
-                                   ret == 0 && ni_req->gethost ? hostbuf : NULL,
-                                   ret == 0 && ni_req->getserv ? servbuf : NULL);
-    }
+            return send_nameinfo_reply(out_fd, req->id, ret,
+                                       ret == 0 && ni_req->gethost ? hostbuf : NULL,
+                                       ret == 0 && ni_req->getserv ? servbuf : NULL,
+                                       errno, h_errno);
+        }
 
-    case REQUEST_RES_QUERY: 
-    case REQUEST_RES_SEARCH: {
-        int ret;
-        unsigned char answer[BUFSIZE];
-        const res_request_t *res_req = (const res_request_t *)req;
-        const char *dname;
-
-        assert(length >= sizeof(res_request_t));
-        assert(length == sizeof(res_request_t) + res_req->dlen + 1);
+        case REQUEST_RES_QUERY:
+        case REQUEST_RES_SEARCH: {
+            int ret;
+            HEADER answer[BUFSIZE/sizeof(HEADER) + 1];
+            const res_request_t *res_req = (const res_request_t *)req;
+            const char *dname;
 
-        dname = (const char *) req + sizeof(res_request_t);
+            assert(length >= sizeof(res_request_t));
+            assert(length == sizeof(res_request_t) + res_req->dname_len);
+
+            dname = (const char *) req + sizeof(res_request_t);
 
-        if (req->type == REQUEST_RES_QUERY) { 
-            /*printf("[res query for '%s']\n", dname);*/
-            ret = res_query(dname, res_req->class, res_req->type, 
-                            answer, BUFSIZE);
-            /*printf("[return value: %d]\n", ret);*/
-        } else {
-            ret = res_search(dname, res_req->class, res_req->type, 
-                             answer, BUFSIZE);
+            if (req->type == REQUEST_RES_QUERY)
+                ret = res_query(dname, res_req->class, res_req->type, (unsigned char *) answer, BUFSIZE);
+            else
+                ret = res_search(dname, res_req->class, res_req->type, (unsigned char *) answer, BUFSIZE);
+
+            return send_res_reply(out_fd, req->id, (unsigned char *) answer, ret, errno, h_errno);
         }
-        return send_res_reply(out_fd, req->id, answer, ret);
-    }
 
-    case REQUEST_TERMINATE: {
-        /* Quit */
-        return -1;
-    }
+        case REQUEST_TERMINATE:
+            /* Quit */
+            return -1;
 
-    default:
-        ;
+        default:
+            ;
     }
 
     return 0;
@@ -381,41 +590,64 @@
 
 static int process_worker(int in_fd, int out_fd) {
     int have_death_sig = 0;
+    int good_fds[3];
+    int ret = 1;
+
+    const int ignore_sigs[] = {
+        SIGINT,
+        SIGHUP,
+        SIGPIPE,
+        SIGUSR1,
+        SIGUSR2,
+        -1
+    };
+
     assert(in_fd > 2);
     assert(out_fd > 2);
-    
+
     close(0);
     close(1);
     close(2);
 
-    open("/dev/null", O_RDONLY);
-    open("/dev/null", O_WRONLY);
-    open("/dev/null", O_WRONLY);
+    if (open("/dev/null", O_RDONLY) != 0)
+        goto fail;
+
+    if (open("/dev/null", O_WRONLY) != 1)
+        goto fail;
 
-    chdir("/");
+    if (open("/dev/null", O_WRONLY) != 2)
+        goto fail;
+
+    if (chdir("/") < 0)
+        goto fail;
 
     if (geteuid() == 0) {
         struct passwd *pw;
+        int r;
 
         if ((pw = getpwnam("nobody"))) {
-#ifdef HAVE_SETRESUID            
-            setresuid(pw->pw_uid, pw->pw_uid, pw->pw_uid);
+#ifdef HAVE_SETRESUID
+            r = setresuid(pw->pw_uid, pw->pw_uid, pw->pw_uid);
 #elif HAVE_SETREUID
-            setreuid(pw->pw_uid, pw->pw_uid);
+            r = setreuid(pw->pw_uid, pw->pw_uid);
 #else
-            setuid(pw->pw_uid);
-            seteuid(pw->pw_uid);
+            if ((r = setuid(pw->pw_uid)) >= 0)
+                r = seteuid(pw->pw_uid);
 #endif
+            if (r < 0)
+                goto fail;
         }
     }
 
-    signal(SIGTERM, SIG_DFL);
+    if (reset_sigsv(ignore_sigs) < 0)
+        goto fail;
 
-    signal(SIGINT, SIG_IGN);
-    signal(SIGHUP, SIG_IGN);
-    signal(SIGPIPE, SIG_IGN);
-    signal(SIGUSR1, SIG_IGN);
-    signal(SIGUSR2, SIG_IGN);
+    if (ignore_sigsv(ignore_sigs) < 0)
+        goto fail;
+
+    good_fds[0] = in_fd; good_fds[1] = out_fd; good_fds[2] = -1;
+    if (close_allv(good_fds) < 0)
+        goto fail;
 
 #ifdef PR_SET_PDEATHSIG
     if (prctl(PR_SET_PDEATHSIG, SIGTERM) >= 0)
@@ -424,25 +656,25 @@
 
     if (!have_death_sig)
         fd_nonblock(in_fd);
-    
+
     while (getppid() > 1) { /* if the parent PID is 1 our parent process died. */
-        char buf[BUFSIZE];
+        rheader_t buf[BUFSIZE/sizeof(rheader_t) + 1];
         ssize_t length;
 
         if (!have_death_sig) {
             fd_set fds;
-            struct timeval tv = { 0, 500000 } ;
-            
+            struct timeval tv = { 0, 500000 };
+
             FD_ZERO(&fds);
             FD_SET(in_fd, &fds);
-            
+
             if (select(in_fd+1, &fds, NULL, NULL, &tv) < 0)
                 break;
-            
+
             if (getppid() == 1)
                 break;
         }
-        
+
         if ((length = recv(in_fd, buf, sizeof(buf), 0)) <= 0) {
 
             if (length < 0 && errno == EAGAIN)
@@ -451,14 +683,17 @@
             break;
         }
 
-        if (handle_request(out_fd, (rheader_t*) buf, (size_t) length) < 0)
+        if (handle_request(out_fd, buf, (size_t) length) < 0)
             break;
     }
 
-    close(in_fd);
-    close(out_fd);
-    
-    return 0;
+    ret = 0;
+
+fail:
+
+    send_died(out_fd);
+
+    return ret;
 }
 
 #else
@@ -466,24 +701,29 @@
 static void* thread_worker(void *p) {
     sigset_t fullset;
     int *fds = p;
-    pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
+    int in_fd, out_fd;
+
+    in_fd = fds[REQUEST_RECV_FD];
+    out_fd = fds[RESPONSE_SEND_FD];
+    free(p);
 
     /* No signals in this thread please */
     sigfillset(&fullset);
     pthread_sigmask(SIG_BLOCK, &fullset, NULL);
-    
+
     for (;;) {
-        char buf[BUFSIZE];
+        rheader_t buf[BUFSIZE/sizeof(rheader_t) + 1];
         ssize_t length;
 
-        if ((length = recv(fds[REQUEST_RECV_FD], buf, sizeof(buf), 0)) <= 0)
+        if ((length = recv(in_fd, buf, sizeof(buf), 0)) <= 0)
             break;
 
-        if (handle_request(fds[RESPONSE_SEND_FD], (rheader_t*) buf, (size_t) length) < 0)
+        if (handle_request(out_fd, buf, (size_t) length) < 0)
             break;
-        
     }
 
+    send_died(out_fd);
+
     return NULL;
 }
 
@@ -492,28 +732,29 @@
 asyncns_t* asyncns_new(unsigned n_proc) {
     asyncns_t *asyncns = NULL;
     int i;
-    unsigned p;
     assert(n_proc >= 1);
 
     if (n_proc > MAX_WORKERS)
         n_proc = MAX_WORKERS;
 
-    if (!(asyncns = malloc(sizeof(asyncns_t))))
+    if (!(asyncns = malloc(sizeof(asyncns_t)))) {
+        errno = ENOMEM;
         goto fail;
+    }
 
+    asyncns->dead = 0;
     asyncns->valid_workers = 0;
 
-    for (i = 0; i < 4; i++) 
+    for (i = 0; i < MESSAGE_FD_MAX; i++)
         asyncns->fds[i] = -1;
-        
-    for (p = 0; p < MAX_QUERIES; p++)
-        asyncns->queries[p] = NULL;
-    
+
+    memset(asyncns->queries, 0, sizeof(asyncns->queries));
+
     if (socketpair(PF_UNIX, SOCK_DGRAM, 0, asyncns->fds) < 0 ||
         socketpair(PF_UNIX, SOCK_DGRAM, 0, asyncns->fds+2) < 0)
         goto fail;
-    
-    for (i = 0; i < 4; i++) 
+
+    for (i = 0; i < MESSAGE_FD_MAX; i++)
         fd_cloexec(asyncns->fds[i]);
 
     for (asyncns->valid_workers = 0; asyncns->valid_workers < n_proc; asyncns->valid_workers++) {
@@ -522,13 +763,35 @@
         if ((asyncns->workers[asyncns->valid_workers] = fork()) < 0)
             goto fail;
         else if (asyncns->workers[asyncns->valid_workers] == 0) {
+            int ret;
+
             close(asyncns->fds[REQUEST_SEND_FD]);
             close(asyncns->fds[RESPONSE_RECV_FD]);
-            _exit(process_worker(asyncns->fds[REQUEST_RECV_FD], asyncns->fds[RESPONSE_SEND_FD]));
+            ret = process_worker(asyncns->fds[REQUEST_RECV_FD], asyncns->fds[RESPONSE_SEND_FD]);
+            close(asyncns->fds[REQUEST_RECV_FD]);
+            close(asyncns->fds[RESPONSE_SEND_FD]);
+            _exit(ret);
         }
 #else
-        if (pthread_create(&asyncns->workers[asyncns->valid_workers], NULL, thread_worker, asyncns->fds) != 0)
+
+        int *fds, r;
+
+        /* We need to copy this array because otherwise we might have
+         * a small chance of a race where the thread accesses fds when
+         * *asyncns is already dead */
+
+        if (!(fds = malloc(sizeof(asyncns->fds)))) {
+            errno = ENOMEM;
             goto fail;
+        }
+
+        memcpy(fds, asyncns->fds, sizeof(asyncns->fds));
+
+        if ((r = pthread_create(&asyncns->workers[asyncns->valid_workers], NULL, thread_worker, fds)) != 0) {
+            free(fds);
+            errno = r;
+            goto fail;
+        }
 #endif
     }
 
@@ -536,8 +799,8 @@
     close(asyncns->fds[REQUEST_RECV_FD]);
     close(asyncns->fds[RESPONSE_SEND_FD]);
     asyncns->fds[REQUEST_RECV_FD] = asyncns->fds[RESPONSE_SEND_FD] = -1;
-#endif    
-    
+#endif
+
     asyncns->current_index = asyncns->current_id = 0;
     asyncns->done_head = asyncns->done_tail = NULL;
     asyncns->n_queries = 0;
@@ -546,51 +809,61 @@
 
     return asyncns;
 
- fail:
-    if (asyncns) 
+fail:
+    if (asyncns)
         asyncns_free(asyncns);
 
     return NULL;
 }
 
 void asyncns_free(asyncns_t *asyncns) {
+    int i;
+    int saved_errno = errno;
     unsigned p;
-    int i;
-    rheader_t req;
+
     assert(asyncns);
 
-    req.type = REQUEST_TERMINATE;
-    req.length = sizeof(req);
-    req.id = 0;
-    
-    /* Send one termiantion packet for each worker */
-    for (p = 0; p < asyncns->valid_workers; p++)
-        send(asyncns->fds[REQUEST_SEND_FD], &req, req.length, 0);
+    if (asyncns->fds[REQUEST_SEND_FD] >= 0) {
+        rheader_t req;
+
+        memset(&req, 0, sizeof(req));
+        req.type = REQUEST_TERMINATE;
+        req.length = sizeof(req);
+        req.id = 0;
 
-    /* No terminate them forcibly*/
+        /* Send one termiantion packet for each worker */
+        for (p = 0; p < asyncns->valid_workers; p++)
+            send(asyncns->fds[REQUEST_SEND_FD], &req, req.length, 0);
+    }
+
+    /* Close all communication channels */
+    for (i = 0; i < MESSAGE_FD_MAX; i++)
+        if (asyncns->fds[i] >= 0)
+            close(asyncns->fds[i]);
+
+    /* Now terminate them forcibly */
     for (p = 0; p < asyncns->valid_workers; p++) {
 #ifndef HAVE_PTHREAD
         kill(asyncns->workers[p], SIGTERM);
         waitpid(asyncns->workers[p], NULL, 0);
 #else
-        pthread_cancel(asyncns->workers[p]);
-        pthread_join(asyncns->workers[p], NULL);
-#endif        
+        pthread_detach(asyncns->workers[p]);
+
+        /* We don't join the thread here because there is no clean way
+           to cancel a running lookup if one should be active. So it
+           might take a while until the lookup thread actually
+           terminates. But we don't really care, because it won't leak
+           resources. */
+#endif
     }
 
-    /* Due to Solaris' broken thread cancelation we first send an
-     * termination request and then cancel th thread. */
-
-    
-    for (i = 0; i < 4; i++)
-        if (asyncns->fds[i] >= 0)
-            close(asyncns->fds[i]);
-    
     for (p = 0; p < MAX_QUERIES; p++)
         if (asyncns->queries[p])
             asyncns_cancel(asyncns, asyncns->queries[p]);
-    
+
     free(asyncns);
+
+    errno = saved_errno;
 }
 
 int asyncns_fd(asyncns_t *asyncns) {
@@ -616,7 +889,7 @@
     assert(!q->done);
 
     q->done = 1;
-    
+
     if ((q->done_prev = asyncns->done_tail))
         asyncns->done_tail->done_next = q;
     else
@@ -627,7 +900,7 @@
 }
 
 static void *unserialize_addrinfo(void *p, struct addrinfo **ret_ai, size_t *length) {
-    addrinfo_serialization_t *s = p;
+    addrinfo_serialization_t s;
     size_t l;
     struct addrinfo *ai;
     assert(p);
@@ -637,42 +910,44 @@
     if (*length < sizeof(addrinfo_serialization_t))
         return NULL;
 
-    l = sizeof(addrinfo_serialization_t) + s->ai_addrlen + s->canonname_len;
+    memcpy(&s, p, sizeof(s));
+
+    l = sizeof(addrinfo_serialization_t) + s.ai_addrlen + s.canonname_len;
     if (*length < l)
         return NULL;
 
     if (!(ai = malloc(sizeof(struct addrinfo))))
         goto fail;
-    
+
     ai->ai_addr = NULL;
     ai->ai_canonname = NULL;
     ai->ai_next = NULL;
 
-    if (s->ai_addrlen && !(ai->ai_addr = malloc(s->ai_addrlen)))
+    if (s.ai_addrlen && !(ai->ai_addr = malloc(s.ai_addrlen)))
         goto fail;
-    
-    if (s->canonname_len && !(ai->ai_canonname = malloc(s->canonname_len)))
+
+    if (s.canonname_len && !(ai->ai_canonname = malloc(s.canonname_len)))
         goto fail;
 
-    ai->ai_flags = s->ai_flags;
-    ai->ai_family = s->ai_family;
-    ai->ai_socktype = s->ai_socktype;
-    ai->ai_protocol = s->ai_protocol;
-    ai->ai_addrlen = s->ai_addrlen;
+    ai->ai_flags = s.ai_flags;
+    ai->ai_family = s.ai_family;
+    ai->ai_socktype = s.ai_socktype;
+    ai->ai_protocol = s.ai_protocol;
+    ai->ai_addrlen = s.ai_addrlen;
 
     if (ai->ai_addr)
-        memcpy(ai->ai_addr, (uint8_t*) p + sizeof(addrinfo_serialization_t), s->ai_addrlen);
+        memcpy(ai->ai_addr, (uint8_t*) p + sizeof(addrinfo_serialization_t), s.ai_addrlen);
 
     if (ai->ai_canonname)
-        memcpy(ai->ai_canonname, (uint8_t*) p + sizeof(addrinfo_serialization_t) + s->ai_addrlen, s->canonname_len);
+        memcpy(ai->ai_canonname, (uint8_t*) p + sizeof(addrinfo_serialization_t) + s.ai_addrlen, s.canonname_len);
 
     *length -= l;
     *ret_ai = ai;
-    
+
     return (uint8_t*) p + l;
 
 
- fail:
+fail:
     if (ai)
         asyncns_freeaddrinfo(ai);
 
@@ -686,82 +961,99 @@
     assert(length >= sizeof(rheader_t));
     assert(length == resp->length);
 
-    if (!(q = lookup_query(asyncns, resp->id)))
+    if (resp->type == RESPONSE_DIED) {
+        asyncns->dead = 1;
         return 0;
-    
-    switch (resp->type) {
-    case RESPONSE_ADDRINFO: {
-        const addrinfo_response_t *ai_resp = (addrinfo_response_t*) resp;
-        void *p;
-        size_t l;
-        struct addrinfo *prev = NULL;
-
-        assert(length >= sizeof(addrinfo_response_t));
-        assert(q->type == REQUEST_ADDRINFO);
-
-        q->ret = ai_resp->ret;
-        l = length - sizeof(addrinfo_response_t);
-        p = (uint8_t*) resp + sizeof(addrinfo_response_t);
-
-        while (l > 0 && p) {
-            struct addrinfo *ai = NULL;
-            p = unserialize_addrinfo(p, &ai, &l);
-
-            if (!ai)
-                break;
-
-            if (prev)
-                prev->ai_next = ai;
-            else
-                q->addrinfo = ai;
-
-            prev = ai;
-        }
-
-        complete_query(asyncns, q);
-        break;
     }
 
-    case RESPONSE_NAMEINFO: {
-        const nameinfo_response_t *ni_resp = (nameinfo_response_t*) resp;
+    if (!(q = lookup_query(asyncns, resp->id)))
+        return 0;
+
+    switch (resp->type) {
+        case RESPONSE_ADDRINFO: {
+            const addrinfo_response_t *ai_resp = (addrinfo_response_t*) resp;
+            void *p;
+            size_t l;
+            struct addrinfo *prev = NULL;
+
+            assert(length >= sizeof(addrinfo_response_t));
+            assert(q->type == REQUEST_ADDRINFO);
+
+            q->ret = ai_resp->ret;
+            q->_errno = ai_resp->_errno;
+            q->_h_errno = ai_resp->_h_errno;
+            l = length - sizeof(addrinfo_response_t);
+            p = (uint8_t*) resp + sizeof(addrinfo_response_t);
 
-        assert(length >= sizeof(nameinfo_response_t));
-        assert(q->type == REQUEST_NAMEINFO);
+            while (l > 0 && p) {
+                struct addrinfo *ai = NULL;
+                p = unserialize_addrinfo(p, &ai, &l);
+
+                if (!p || !ai) {
+                    q->ret = EAI_MEMORY;
+                    break;
+                }
 
-        q->ret = ni_resp->ret;
+                if (prev)
+                    prev->ai_next = ai;
+                else
+                    q->addrinfo = ai;
+
+                prev = ai;
+            }
+
+            complete_query(asyncns, q);
+            break;
+        }
+
+        case RESPONSE_NAMEINFO: {
+            const nameinfo_response_t *ni_resp = (nameinfo_response_t*) resp;
 
-        if (ni_resp->hostlen)
-            q->host = g_strndup((const char*) ni_resp + sizeof(nameinfo_response_t), (gsize)ni_resp->hostlen-1);
+            assert(length >= sizeof(nameinfo_response_t));
+            assert(q->type == REQUEST_NAMEINFO);
+
+            q->ret = ni_resp->ret;
+            q->_errno = ni_resp->_errno;
+            q->_h_errno = ni_resp->_h_errno;
+
+            if (ni_resp->hostlen)
+                if (!(q->host = strndup((const char*) ni_resp + sizeof(nameinfo_response_t), ni_resp->hostlen-1)))
+                    q->ret = EAI_MEMORY;
+
+            if (ni_resp->servlen)
+                if (!(q->serv = strndup((const char*) ni_resp + sizeof(nameinfo_response_t) + ni_resp->hostlen, ni_resp->servlen-1)))
+                    q->ret = EAI_MEMORY;
+
+            complete_query(asyncns, q);
+            break;
+        }
+
+        case RESPONSE_RES: {
+            const res_response_t *res_resp = (res_response_t *)resp;
 
-        if (ni_resp->servlen)
-            q->serv = g_strndup((const char*) ni_resp + sizeof(nameinfo_response_t) + ni_resp->hostlen, (gsize)ni_resp->servlen-1);
-                    
+            assert(length >= sizeof(res_response_t));
+            assert(q->type == REQUEST_RES_QUERY || q->type == REQUEST_RES_SEARCH);
+
+            q->ret = res_resp->ret;
+            q->_errno = res_resp->_errno;
+            q->_h_errno = res_resp->_h_errno;
 
-        complete_query(asyncns, q);
-        break;
+            if (res_resp->ret >= 0)  {
+                if (!(q->serv = malloc(res_resp->ret))) {
+                    q->ret = -1;
+                    q->_errno = ENOMEM;
+                } else
+                    memcpy(q->serv, (char *)resp + sizeof(res_response_t), res_resp->ret);
+            }
+
+            complete_query(asyncns, q);
+            break;
+        }
+
+        default:
+            ;
     }
 
-    case RESPONSE_RES: {
-        const res_response_t *res_resp = (res_response_t *)resp;
-
-        assert(length >= sizeof(res_response_t));
-        assert(q->type == REQUEST_RES_QUERY || q->type == REQUEST_RES_SEARCH);
-
-        q->ret = res_resp->ret;
-
-        if (res_resp->ret >= 0)  {
-            q->serv = malloc(res_resp->ret);
-            memcpy(q->serv, (char *)resp + sizeof(res_response_t), res_resp->ret);
-        }
-
-        complete_query(asyncns, q);
-        break;
-    }
-            
-    default:
-        ;
-    }
-    
     return 0;
 }
 
@@ -770,9 +1062,14 @@
     assert(asyncns);
 
     for (;;) {
-        char buf[BUFSIZE];
+        rheader_t buf[BUFSIZE/sizeof(rheader_t) + 1];
         ssize_t l;
 
+        if (asyncns->dead) {
+            errno = ECHILD;
+            return -1;
+        }
+
         if (((l = recv(asyncns->fds[RESPONSE_RECV_FD], buf, sizeof(buf), 0)) < 0)) {
             fd_set fds;
 
@@ -781,18 +1078,17 @@
 
             if (!block || handled)
                 return 0;
-                
+
             FD_ZERO(&fds);
             FD_SET(asyncns->fds[RESPONSE_RECV_FD], &fds);
 
             if (select(asyncns->fds[RESPONSE_RECV_FD]+1, &fds, NULL, NULL, NULL) < 0)
                 return -1;
-                
+
             continue;
         }
 
-
-        if (handle_response(asyncns, (rheader_t*) buf, (size_t) l) < 0)
+        if (handle_response(asyncns, buf, (size_t) l) < 0)
             return -1;
 
         handled = 1;
@@ -803,8 +1099,10 @@
     asyncns_query_t *q;
     assert(asyncns);
 
-    if (asyncns->n_queries >= MAX_QUERIES)
+    if (asyncns->n_queries >= MAX_QUERIES) {
+        errno = ENOMEM;
         return NULL;
+    }
 
     while (asyncns->queries[asyncns->current_index]) {
 
@@ -814,9 +1112,11 @@
         while (asyncns->current_index >= MAX_QUERIES)
             asyncns->current_index -= MAX_QUERIES;
     }
-        
-    if (!(q = asyncns->queries[asyncns->current_index] = malloc(sizeof(asyncns_query_t))))
+
+    if (!(q = asyncns->queries[asyncns->current_index] = malloc(sizeof(asyncns_query_t)))) {
+        errno = ENOMEM;
         return NULL;
+    }
 
     asyncns->n_queries++;
 
@@ -825,6 +1125,8 @@
     q->id = asyncns->current_id;
     q->done_next = q->done_prev = NULL;
     q->ret = 0;
+    q->_errno = 0;
+    q->_h_errno = 0;
     q->addrinfo = NULL;
     q->userdata = NULL;
     q->host = q->serv = NULL;
@@ -833,29 +1135,36 @@
 }
 
 asyncns_query_t* asyncns_getaddrinfo(asyncns_t *asyncns, const char *node, const char *service, const struct addrinfo *hints) {
-    uint8_t data[BUFSIZE];
-    addrinfo_request_t *req = (addrinfo_request_t*) data;
+    addrinfo_request_t data[BUFSIZE/sizeof(addrinfo_request_t) + 1];
+    addrinfo_request_t *req = data;
     asyncns_query_t *q;
     assert(asyncns);
     assert(node || service);
 
+    if (asyncns->dead) {
+        errno = ECHILD;
+        return NULL;
+    }
+
     if (!(q = alloc_query(asyncns)))
         return NULL;
 
     memset(req, 0, sizeof(addrinfo_request_t));
-    
+
     req->node_len = node ? strlen(node)+1 : 0;
     req->service_len = service ? strlen(service)+1 : 0;
-    
+
     req->header.id = q->id;
     req->header.type = q->type = REQUEST_ADDRINFO;
     req->header.length = sizeof(addrinfo_request_t) + req->node_len + req->service_len;
 
-    if (req->header.length > BUFSIZE)
+    if (req->header.length > BUFSIZE) {
+        errno = ENOMEM;
         goto fail;
+    }
 
     if (!(req->hints_is_null = !hints)) {
-        req->ai_flags = hints->ai_flags; 
+        req->ai_flags = hints->ai_flags;
         req->ai_family = hints->ai_family;
         req->ai_socktype = hints->ai_socktype;
         req->ai_protocol = hints->ai_protocol;
@@ -869,10 +1178,10 @@
 
     if (send(asyncns->fds[REQUEST_SEND_FD], req, req->header.length, 0) < 0)
         goto fail;
-    
+
     return q;
 
- fail:
+fail:
     if (q)
         asyncns_cancel(asyncns, q);
 
@@ -886,37 +1195,57 @@
     assert(q->asyncns == asyncns);
     assert(q->type == REQUEST_ADDRINFO);
 
+    if (asyncns->dead) {
+        errno = ECHILD;
+        return EAI_SYSTEM;
+    }
+
     if (!q->done)
         return EAI_AGAIN;
 
     *ret_res = q->addrinfo;
     q->addrinfo = NULL;
+
     ret = q->ret;
+
+    if (ret == EAI_SYSTEM)
+        errno = q->_errno;
+
+    if (ret != 0)
+        h_errno = q->_h_errno;
+
     asyncns_cancel(asyncns, q);
 
     return ret;
 }
 
 asyncns_query_t* asyncns_getnameinfo(asyncns_t *asyncns, const struct sockaddr *sa, socklen_t salen, int flags, int gethost, int getserv) {
-    uint8_t data[BUFSIZE];
-    nameinfo_request_t *req = (nameinfo_request_t*) data;
+    nameinfo_request_t data[BUFSIZE/sizeof(nameinfo_request_t) + 1];
+    nameinfo_request_t *req = data;
     asyncns_query_t *q;
 
     assert(asyncns);
     assert(sa);
     assert(salen > 0);
 
+    if (asyncns->dead) {
+        errno = ECHILD;
+        return NULL;
+    }
+
     if (!(q = alloc_query(asyncns)))
         return NULL;
 
     memset(req, 0, sizeof(nameinfo_request_t));
-    
+
     req->header.id = q->id;
     req->header.type = q->type = REQUEST_NAMEINFO;
     req->header.length = sizeof(nameinfo_request_t) + salen;
 
-    if (req->header.length > BUFSIZE)
+    if (req->header.length > BUFSIZE) {
+        errno = ENOMEM;
         goto fail;
+    }
 
     req->flags = flags;
     req->sockaddr_len = salen;
@@ -927,10 +1256,10 @@
 
     if (send(asyncns->fds[REQUEST_SEND_FD], req, req->header.length, 0) < 0)
         goto fail;
-    
+
     return q;
 
- fail:
+fail:
     if (q)
         asyncns_cancel(asyncns, q);
 
@@ -946,6 +1275,11 @@
     assert(!ret_host || hostlen);
     assert(!ret_serv || servlen);
 
+    if (asyncns->dead) {
+        errno = ECHILD;
+        return EAI_SYSTEM;
+    }
+
     if (!q->done)
         return EAI_AGAIN;
 
@@ -958,61 +1292,71 @@
         strncpy(ret_serv, q->serv, servlen);
         ret_serv[servlen-1] = 0;
     }
-    
+
     ret = q->ret;
+
+    if (ret == EAI_SYSTEM)
+        errno = q->_errno;
+
+    if (ret != 0)
+        h_errno = q->_h_errno;
+
     asyncns_cancel(asyncns, q);
 
     return ret;
 }
 
-static asyncns_query_t *
-asyncns_res(asyncns_t *asyncns, query_type_t qtype, 
-            const char *dname, int class, int type) {
-    uint8_t data[BUFSIZE];
-    res_request_t *req = (res_request_t*) data;
+static asyncns_query_t * asyncns_res(asyncns_t *asyncns, query_type_t qtype, const char *dname, int class, int type) {
+    res_request_t data[BUFSIZE/sizeof(res_request_t) + 1];
+    res_request_t *req = data;
     asyncns_query_t *q;
-    size_t dlen;
 
     assert(asyncns);
     assert(dname);
 
-    dlen = strlen(dname);
+    if (asyncns->dead) {
+        errno = ECHILD;
+        return NULL;
+    }
 
     if (!(q = alloc_query(asyncns)))
         return NULL;
 
     memset(req, 0, sizeof(res_request_t));
-    
+
+    req->dname_len = strlen(dname) + 1;
+
     req->header.id = q->id;
     req->header.type = q->type = qtype;
-    req->header.length = sizeof(res_request_t) + dlen + 1;
+    req->header.length = sizeof(res_request_t) + req->dname_len;
 
-    if (req->header.length > BUFSIZE)
+    if (req->header.length > BUFSIZE) {
+        errno = ENOMEM;
         goto fail;
+    }
 
     req->class = class;
     req->type = type;
-    req->dlen = dlen;
 
-    memcpy((uint8_t*) req + sizeof(res_request_t), dname, dlen + 1);
+    strcpy((char*) req + sizeof(res_request_t), dname);
 
     if (send(asyncns->fds[REQUEST_SEND_FD], req, req->header.length, 0) < 0)
         goto fail;
 
     return q;
 
- fail:
+fail:
     if (q)
         asyncns_cancel(asyncns, q);
 
     return NULL;
 }
 
-asyncns_query_t* asyncns_res_query(asyncns_t *asyncns, const char *dname, int class, int type) { 
+asyncns_query_t* asyncns_res_query(asyncns_t *asyncns, const char *dname, int class, int type) {
     return asyncns_res(asyncns, REQUEST_RES_QUERY, dname, class, type);
 }
 
-asyncns_query_t* asyncns_res_search(asyncns_t *asyncns, const char *dname, int class, int type) { 
+asyncns_query_t* asyncns_res_search(asyncns_t *asyncns, const char *dname, int class, int type) {
     return asyncns_res(asyncns, REQUEST_RES_SEARCH, dname, class, type);
 }
 
@@ -1024,17 +1368,29 @@
     assert(q->type == REQUEST_RES_QUERY || q->type == REQUEST_RES_SEARCH);
     assert(answer);
 
-    if (!q->done)
+    if (asyncns->dead) {
+        errno = ECHILD;
+        return -ECHILD;
+    }
+
+    if (!q->done) {
+        errno = EAGAIN;
         return -EAGAIN;
+    }
 
     *answer = (unsigned char *)q->serv;
     q->serv = NULL;
 
     ret = q->ret;
 
+    if (ret < 0) {
+        errno = q->_errno;
+        h_errno = q->_h_errno;
+    }
+
     asyncns_cancel(asyncns, q);
 
-    return ret;
+    return ret < 0 ? -errno : ret;
 }
 
 asyncns_query_t* asyncns_getnext(asyncns_t *asyncns) {
@@ -1049,6 +1405,8 @@
 
 void asyncns_cancel(asyncns_t *asyncns, asyncns_query_t* q) {
     int i;
+    int saved_errno = errno;
+
     assert(asyncns);
     assert(q);
     assert(q->asyncns == asyncns);
@@ -1058,7 +1416,7 @@
 
         if (q->done_prev)
             q->done_prev->done_next = q->done_next;
-        else 
+        else
             asyncns->done_head = q->done_next;
 
         if (q->done_next)
@@ -1067,23 +1425,22 @@
             asyncns->done_tail = q->done_prev;
     }
 
-
     i = q->id % MAX_QUERIES;
     assert(asyncns->queries[i] == q);
     asyncns->queries[i] = NULL;
 
     asyncns_freeaddrinfo(q->addrinfo);
-    free(q->addrinfo);
-    g_free(q->host);
-    g_free(q->serv);
+    free(q->host);
+    free(q->serv);
 
     asyncns->n_queries--;
     free(q);
+
+    errno = saved_errno;
 }
 
 void asyncns_freeaddrinfo(struct addrinfo *ai) {
-    if (!ai)
-        return;
+    int saved_errno = errno;
 
     while (ai) {
         struct addrinfo *next = ai->ai_next;
@@ -1094,6 +1451,24 @@
 
         ai = next;
     }
+
+    errno = saved_errno;
+}
+
+void asyncns_freeanswer(unsigned char *answer) {
+    int saved_errno = errno;
+
+    if (!answer)
+        return;
+
+    /* Please note that this function is new in libasyncns 0.4. In
+     * older versions you were supposed to free the answer directly
+     * with free(). Hence, if this function is changed to do more than
+     * just a simple free() this must be considered ABI/API breakage! */
+
+    free(answer);
+
+    errno = saved_errno;
 }
 
 int asyncns_isdone(asyncns_t *asyncns, asyncns_query_t*q) {
@@ -1108,7 +1483,7 @@
     assert(q);
     assert(asyncns);
     assert(q->asyncns = asyncns);
-    
+
     q->userdata = userdata;
 }
 
@@ -1119,5 +1494,3 @@
 
     return q->userdata;
 }
-
-
--- a/loudmouth/lm-old-socket.c	Sat Oct 25 22:33:23 2008 +0200
+++ b/loudmouth/lm-old-socket.c	Sun Oct 26 01:52:23 2008 +0200
@@ -820,7 +820,6 @@
     lm_resolver_lookup (socket->resolver);
 }
 
-
 LmOldSocket *
 lm_old_socket_create (GMainContext      *context,
                       IncomingDataFunc   data_func,
--- a/loudmouth/lm-resolver.c	Sat Oct 25 22:33:23 2008 +0200
+++ b/loudmouth/lm-resolver.c	Sun Oct 26 01:52:23 2008 +0200
@@ -32,6 +32,7 @@
 
 #include "lm-asyncns-resolver.h"
 #include "lm-blocking-resolver.h"
+#include "lm-debug.h"
 #include "lm-internals.h"
 #include "lm-marshal.h"
 #include "lm-resolver.h"
--- a/loudmouth/asyncns.h	Sat Oct 25 22:33:23 2008 +0200
+++ b/loudmouth/asyncns.h	Sun Oct 26 01:52:23 2008 +0200
@@ -1,26 +1,24 @@
-/* -*- mode: c; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
 #ifndef fooasyncnshfoo
 #define fooasyncnshfoo
 
-/* $Id: asyncns.h 23 2007-02-16 12:49:17Z lennart $ */
+/***
+  This file is part of libasyncns.
+
+  Copyright 2005-2008 Lennart Poettering
 
-/***
-    This file is part of libasyncns.
- 
-    libasyncns is free software; you can redistribute it and/or modify
-    it under the terms of the GNU Lesser General Public License as
-    published by the Free Software Foundation; either version 2 of the
-    License, or (at your option) any later version.
- 
-    libasyncns is distributed in the hope that it will be useful, but
-    WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
-    General Public License for more details.
- 
-    You should have received a copy of the GNU Lesser General Public
-    License along with libasyncns; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
-    USA.
+  libasyncns is free software; you can redistribute it and/or modify
+  it under the terms of the GNU Lesser General Public License as
+  published by the Free Software Foundation, either version 2.1 of the
+  License, or (at your option) any later version.
+
+  libasyncns is distributed in the hope that it will be useful, but
+  WITHOUT ANY WARRANTY; without even the implied warranty of
+  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
+  Lesser General Public License for more details.
+
+  You should have received a copy of the GNU Lesser General Public
+  License along with libasyncns. If not, see
+  <http://www.gnu.org/licenses/>.
 ***/
 
 #include <sys/types.h>
@@ -50,36 +48,36 @@
 #endif
 
 /** An opaque libasyncns session structure */
-    typedef struct asyncns asyncns_t;
+typedef struct asyncns asyncns_t;
 
 /** An opaque libasyncns query structure */
-    typedef struct asyncns_query asyncns_query_t;
+typedef struct asyncns_query asyncns_query_t;
 
-/** Allocate a new libasyncns session with n_proc worker processes */
-    asyncns_t* asyncns_new(unsigned n_proc);
+/** Allocate a new libasyncns session with n_proc worker processes/threads */
+asyncns_t* asyncns_new(unsigned n_proc);
 
 /** Free a libasyncns session. This destroys all attached
  * asyncns_query_t objects automatically */
-    void asyncns_free(asyncns_t *asyncns);
+void asyncns_free(asyncns_t *asyncns);
 
 /** Return the UNIX file descriptor to select() for readability
  * on. Use this function to integrate libasyncns with your custom main
  * loop. */
-    int asyncns_fd(asyncns_t *asyncns);
+int asyncns_fd(asyncns_t *asyncns);
 
 /** Process pending responses. After this function is called you can
  * get the next completed query object(s) using asyncns_getnext(). If
  * block is non-zero wait until at least one response has been
  * processed. If block is zero, process all pending responses and
  * return. */
-    int asyncns_wait(asyncns_t *asyncns, int block);
+int asyncns_wait(asyncns_t *asyncns, int block);
 
 /** Issue a name to address query on the specified session. The
  * arguments are compatible with the ones of libc's
  * getaddrinfo(3). The function returns a new query object. When the
  * query is completed you may retrieve the results using
  * asyncns_getaddrinfo_done().*/
-    asyncns_query_t* asyncns_getaddrinfo(asyncns_t *asyncns, const char *node, const char *service, const struct addrinfo *hints);
+asyncns_query_t* asyncns_getaddrinfo(asyncns_t *asyncns, const char *node, const char *service, const struct addrinfo *hints);
 
 /** Retrieve the results of a preceding asyncns_getaddrinfo()
  * call. Returns a addrinfo structure and a return value compatible
@@ -88,74 +86,78 @@
  * returned addrinfo structure with asyncns_freeaddrinfo() and not
  * libc's freeaddrinfo(3)! If the query is not completed yet EAI_AGAIN
  * is returned.*/
-    int asyncns_getaddrinfo_done(asyncns_t *asyncns, asyncns_query_t* q, struct addrinfo **ret_res);
+int asyncns_getaddrinfo_done(asyncns_t *asyncns, asyncns_query_t* q, struct addrinfo **ret_res);
 
 /** Issue an address to name query on the specified session. The
-    arguments are compatible with the ones of libc's getnameinfo(3). The
-    function returns a new query object. When the query is completed you
-    may retrieve the results using asyncns_getnameinfo_done(). Set gethost
-    (resp. getserv) to non-zero if you want to query the hostname
-    (resp. the service name). */
-    asyncns_query_t* asyncns_getnameinfo(asyncns_t *asyncns, const struct sockaddr *sa, socklen_t salen, int flags, int gethost, int getserv);
+ * arguments are compatible with the ones of libc's
+ * getnameinfo(3). The function returns a new query object. When the
+ * query is completed you may retrieve the results using
+ * asyncns_getnameinfo_done(). Set gethost (resp. getserv) to non-zero
+ * if you want to query the hostname (resp. the service name). */
+asyncns_query_t* asyncns_getnameinfo(asyncns_t *asyncns, const struct sockaddr *sa, socklen_t salen, int flags, int gethost, int getserv);
 
-/** Retrieve the results of a preceding asyncns_getnameinfo)(
+/** Retrieve the results of a preceding asyncns_getnameinfo()
  * call. Returns the hostname and the service name in ret_host and
  * ret_serv. The query object q is destroyed by this call and may not
  * be used any further. If the query is not completed yet EAI_AGAIN is
  * returned. */
-    int asyncns_getnameinfo_done(asyncns_t *asyncns, asyncns_query_t* q, char *ret_host, size_t hostlen, char *ret_serv, size_t servlen);
+int asyncns_getnameinfo_done(asyncns_t *asyncns, asyncns_query_t* q, char *ret_host, size_t hostlen, char *ret_serv, size_t servlen);
 
-/** Issue an resolver query on the specified session. The arguments are
+/** Issue a resolver query on the specified session. The arguments are
  * compatible with the ones of libc's res_query(3). The function returns a new
  * query object. When the query is completed you may retrieve the results using
  * asyncns_res_done().  */
-    asyncns_query_t* asyncns_res_query(asyncns_t *asyncns, const char *dname, int class, int type);
+asyncns_query_t* asyncns_res_query(asyncns_t *asyncns, const char *dname, int class, int type);
 
 /** Issue an resolver query on the specified session. The arguments are
  * compatible with the ones of libc's res_search(3). The function returns a new
  * query object. When the query is completed you may retrieve the results using
  * asyncns_res_done().  */
-    asyncns_query_t* asyncns_res_search(asyncns_t *asyncns, const char *dname, int class, int type);
+asyncns_query_t* asyncns_res_search(asyncns_t *asyncns, const char *dname, int class, int type);
 
-/** Retrieve the results of a preceding asyncns_res_query)( or
- * asyncns_res_search call.  The query object q is destroyed by this call and
- * may not be used any further. Returns a pointer to the answer of the
- * res_query call. If the query is not completed yet -EAGAIN is returned, on
- * failure -errno is returned otherwise the length of answer is returned. */
-    int asyncns_res_done(asyncns_t *asyncns, asyncns_query_t* q, unsigned char
-                         **answer);
+/** Retrieve the results of a preceding asyncns_res_query() or
+ * asyncns_res_search call.  The query object q is destroyed by this
+ * call and may not be used any further. Returns a pointer to the
+ * answer of the res_query call. If the query is not completed yet
+ * -EAGAIN is returned, on failure -errno is returned, otherwise the
+ * length of answer is returned. Make sure to free the answer is a
+ * call to asyncns_freeanswer(). */
+int asyncns_res_done(asyncns_t *asyncns, asyncns_query_t* q, unsigned char **answer);
 
 /** Return the next completed query object. If no query has been
  * completed yet, return NULL. Please note that you need to run
  * asyncns_wait() before this function will return sensible data.  */
-    asyncns_query_t* asyncns_getnext(asyncns_t *asyncns);
+asyncns_query_t* asyncns_getnext(asyncns_t *asyncns);
 
 /** Return the number of query objects (completed or not) attached to
  * this session */
-    int asyncns_getnqueries(asyncns_t *asyncns);
+int asyncns_getnqueries(asyncns_t *asyncns);
 
 /** Cancel a currently running query. q is is destroyed by this call
  * and may not be used any futher. */
-    void asyncns_cancel(asyncns_t *asyncns, asyncns_query_t* q);
+void asyncns_cancel(asyncns_t *asyncns, asyncns_query_t* q);
 
 /** Free the addrinfo structure as returned by
-    asyncns_getaddrinfo_done(). Make sure to use this functions instead of
-    the libc's freeaddrinfo()! */
-    void asyncns_freeaddrinfo(struct addrinfo *ai);
+ * asyncns_getaddrinfo_done(). Make sure to use this functions instead
+ * of the libc's freeaddrinfo()! */
+void asyncns_freeaddrinfo(struct addrinfo *ai);
+
+/** Free the answer data as returned by asyncns_res_done().*/
+void asyncns_freeanswer(unsigned char *answer);
 
 /** Returns non-zero when the query operation specified by q has been completed */
-    int asyncns_isdone(asyncns_t *asyncns, asyncns_query_t*q);
+int asyncns_isdone(asyncns_t *asyncns, asyncns_query_t*q);
 
 /** Assign some opaque userdata with a query object */
-    void asyncns_setuserdata(asyncns_t *asyncns, asyncns_query_t *q, void *userdata);
+void asyncns_setuserdata(asyncns_t *asyncns, asyncns_query_t *q, void *userdata);
 
 /** Return userdata assigned to a query object. Use
  * asyncns_setuserdata() to set this data. If no data has been set
  * prior to this call it returns NULL. */
-    void* asyncns_getuserdata(asyncns_t *asyncns, asyncns_query_t *q);
+void* asyncns_getuserdata(asyncns_t *asyncns, asyncns_query_t *q);
 
 #ifdef  __cplusplus
 }
 #endif
-    
+
 #endif